#include "X11App.h"

#include <cstdlib>
#include <iostream>
#include <X11/Xutil.h>

////////////////////////////////////////////////////////////////////////////////
// X11Window
////////////////////////////////////////////////////////////////////////////////
X11Window::X11Window(Display * dpy,
		     const std::string & win_name,
		     int x, int y,
		     unsigned int width, unsigned int height) :
   name (win_name),
   win (0)
   {
   int screen_num = DefaultScreen(dpy);

   win = XCreateSimpleWindow(dpy, 
			     RootWindow(dpy, screen_num),
			     x, y, 
			     width, height, 
			     1, 
			     None, None);

   }

////////////////////////////////////////////////////////////////////////////////
// X11App
////////////////////////////////////////////////////////////////////////////////
X11App::X11App(const std::string & appname) :
   main_win (NULL), dpy (NULL), event_mask (0), wm_delete_window (0)
   {

   if ((dpy = XOpenDisplay(NULL)) == NULL) 
      {
      std::cerr << "Unable to connect to display " << XDisplayName(NULL) << std::endl;
      exit(1);
      }
   std::cout << "Display name     : " << XDisplayName(NULL) << std::endl
	     << "Number of screens: " << ScreenCount(dpy) << std::endl
	     << "Default screen   : " << DefaultScreen(dpy) << std::endl
	     << "Protocol Version : " << ProtocolVersion(dpy) << std::endl
	     << "Protocol Revision: " << ProtocolRevision(dpy) << std::endl
	     << "Server Vendor    : " << ServerVendor(dpy) << std::endl
	     << "Vendor Release   : " << VendorRelease(dpy) << std::endl
	     << std::endl
	     << "Display width    : " << DisplayWidth(dpy, DefaultScreen(dpy)) << std::endl
	     << "Display height   : " << DisplayHeight(dpy, DefaultScreen(dpy)) << std::endl
	 ;

   // Create the main window
   main_win = new X11Window(dpy, "main", 0, 0, 
			    DisplayWidth(dpy, DefaultScreen(dpy)) / 2,
			    DisplayHeight(dpy, DefaultScreen(dpy)) / 2);


   // We need to set the window manager hints, size hints, and window name.

   // Even though we are not passing any actual size_hints, the Xlib programming manual says to pass the flags below anyway.
   XSizeHints *size_hints;
   if (!(size_hints = XAllocSizeHints())) 
      {
      std::cerr << "XAllocSizeHints: memory allocation failure" << std::endl;
      exit(1);
      }
   size_hints->flags = PPosition | PSize;


   // Get X Properties for the window name.
   XTextProperty windowName;
   const char * ptr = appname.c_str();
   if (XStringListToTextProperty(const_cast<char **>(&ptr), 1, &windowName) == 0) 
      {
      std::cerr << "XStringListToTextProperty: structure allocation for windowName failed." << std::endl;;
      exit(1);
      }

   XWMHints *wm_hints;
   if (!(wm_hints = XAllocWMHints()))
      {
      std::cerr << "XAllocWMHints: memory allocation failure" << std::endl;
      exit(0);
      }
   wm_hints->initial_state = NormalState;
   wm_hints->input = True;
   wm_hints->flags = StateHint | InputHint;

   XSetWMProperties(dpy, main_win->win, &windowName, NULL, 
    		    NULL, 0, size_hints, wm_hints,
		    NULL);


   // Select the event types we want to receive.
   event_mask = 
	 ExposureMask | StructureNotifyMask |
	 KeyPressMask | KeyReleaseMask | 
	 FocusChangeMask |
	 PointerMotionMask |
	 EnterWindowMask | LeaveWindowMask |
	 ButtonPressMask | ButtonReleaseMask ;
   XSelectInput(dpy, main_win->win, event_mask);


   // Make sure we get delete events from the window manager.
   // "wm_delete_window" is the Atom which corresponds to the delete
   //  window message sent by the window manager. 

   wm_delete_window = XInternAtom (dpy, "WM_DELETE_WINDOW", False);
   XSetWMProtocols(dpy, main_win->win, &wm_delete_window, 1);


   // Map the window.  Remember: this does not display the window immediately.
   // The request is queued until events are read, or we call XFlush or XSync.
   XMapWindow(dpy, main_win->win);
   }

////////////////////////////////////////////////////////////////////////////////
X11App::~X11App()
   {
   if (main_win) { delete main_win; }
   if (dpy) { XCloseDisplay(dpy); }
   }

////////////////////////////////////////////////////////////////////////////////
void X11App::run(void)
   {
   XEvent e;
   KeySym k;
   char key_buffer[8];
   key_buffer[7] = 0;
   XComposeStatus compose_status;

   int done = 0;

   while (!done)
      {
      while (XCheckIfEvent(dpy, &e, 
      			   // *sniff*  My very first lambda function in C++
      			   [&](Display * d, XEvent * e, XPointer a)->Bool { return True; }, 
      			   NULL))
	 {
	 switch (e.type)
	    {

	    ////////////////////////
	    // General window events
	    case Expose:
	       // We'll only redraw the entire window at a time, so unless this is 
	       // the last contiguous expose, don't draw the window.
	       if (e.xexpose.count != 0)
		  break;
	       // Draw the window contents here
	       break;
	    case ClientMessage:
	       if ((Atom)e.xclient.data.l[0] == wm_delete_window)
		  done = 1;
	       break;

	       ///////////////////////
	       // Mouse events
	    case MotionNotify:
	       std::cout << "Mouse at (" << e.xmotion.x << "," << e.xmotion.y << ")" << std::endl;
	       break;
	    case ButtonPress:
	       std::cout << "You pressed button " << e.xbutton.button << 
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;
	    case ButtonRelease:
	       std::cout << "You released button " << e.xbutton.button <<
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;
	    case EnterNotify:
	       std::cout << "Mouse has entered the window" <<
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;
	    case LeaveNotify:
	       std::cout << "Mouse has left the window" << 
		     " at (" << e.xbutton.x << "," << e.xbutton.y << ")" << 
		     " state " << button_state_string(e.xbutton.state ) << std::endl;
	       break;

	       //////////////////
	       // Keyboard Events
	    case FocusIn:
	       // Interesting note: When we regain focus due to a mouse click in the window, 
	       // we first get a LeaveNotify, then the FocusIn, then an EnterNotify, and then
	       // the ButtonPress and ButtonRelease.  So in a real app we might want to have
	       // the FocusIn handler read ahead and handle (discard) the next ButtonPress,
	       // and ButtonRelease so that the click does not have a real effect on the application.
	       std::cout << "Got keyboard focus" << std::endl;
	       break;
	    case FocusOut:
	       std::cout << "Lost keyboard focus" << std::endl;
	       break;
	    case KeyPress:
	       // Doing the lookup like this, using XLookupString, means I get the modified
	       // character back.  Thus type shift+a gives me XK_A instead of XK_a
	       // If I used XKeySym I'd instead get a sequence such as this:
	       // XK_Shift_L followed by XK_a
	       //
	       // In the end, that might be a more useful way to go, but for now I'll keep
	       // things simple.
	       XLookupString(&(e.xkey), key_buffer, 7, &k, &compose_status);
	       switch (k)
		  {
		  case XK_Escape:
		  case XK_q:
		     std::cout << "Goodbye!" << std::endl;
		     done = 1;
		     break;
		  default:
		     std::cout << "You typed " << key_buffer << std::endl;
		     break;
		  }
	    case MappingNotify:
	       // While it is uncommon for the user to change the keyboard mapping while
	       // a program is running, handling this case is so trivial that there is 
	       // simply no excuse to leave it out.
	       XRefreshKeyboardMapping((XMappingEvent *)&e);
	       break;

	    default:
	       break;
	    }
	 }

      // Do idle time things here

      }
   }

////////////////////////////////////////////////////////////////////////////////
std::string X11App::button_state_string(unsigned int state)
   {
   std::string str = "";
   if (state & Button1Mask) { str.append("Button1 "); }
   if (state & Button2Mask) { str.append("Button2 "); }
   if (state & Button3Mask) { str.append("Button3 "); }
   if (state & Button4Mask) { str.append("Button4 "); }
   if (state & Button5Mask) { str.append("Button5 "); }
   if (state & ShiftMask)   { str.append("Shift "); }
   if (state & LockMask)    { str.append("Lock "); }
   if (state & ControlMask) { str.append("Control "); }
   if (state & Mod1Mask)    { str.append("Mod1 "); }
   if (state & Mod2Mask)    { str.append("Mod2 "); }
   if (state & Mod3Mask)    { str.append("Mod3 "); }
   if (state & Mod4Mask)    { str.append("Mod4 "); }
   if (state & Mod5Mask)    { str.append("Mod5 "); }
   size_t i = str.rfind(' '); 
   if (i != str.npos) { str.erase(i, 1); }
   return str;
   }

